@@ -1,9 +1,7 @@
 import h5py
 from contextlib import contextmanager
 from collections import namedtuple, OrderedDict
-from io import StringIO
 from itertools import chain, product
-import logging
 import lrspline as lr
 import numpy as np
 import splipy.io
@@ -11,179 +9,10 @@ from splipy import SplineObject, BSplineBasis
 from splipy.SplineModel import ObjectCatalogue
 import splipy.utils
 import sys
+import treelog as log
 
-
-class G2Object(splipy.io.G2):
-
-    def __init__(self, fstream, mode):
-        self.fstream = fstream
-        self.onlywrite = mode == 'w'
-        super(G2Object, self).__init__('')
-
-    def __enter__(self):
-        return self
-
-
-class Log:
-
-    _indent = 0
-
-    @classmethod
-    def debug(cls, s, *args, **kwargs):
-        return logging.debug(' ' * cls._indent + s, *args, **kwargs)
-
-    @classmethod
-    def info(cls, s, *args, **kwargs):
-        return logging.info(' ' * cls._indent + s, *args, **kwargs)
-
-    @classmethod
-    def warning(cls, s, *args, **kwargs):
-        return logging.warning(' ' * cls._indent + s, *args, **kwargs)
-
-    @classmethod
-    def error(cls, s, *args, **kwargs):
-        return logging.error(' ' * cls._indent + s, *args, **kwargs)
-
-    @classmethod
-    def critical(cls, s, *args, **kwargs):
-        return logging.critical(' ' * cls._indent + s, *args, **kwargs)
-
-    @classmethod
-    @contextmanager
-    def indent(cls):
-        cls._indent += 4
-        yield
-        cls._indent -= 4
-
-
-class TensorTesselation:
-
-    def __init__(self, patch):
-        self.knots = patch.knots()
-
-    def __call__(self, patch, coeffs=None, cells=False):
-        if cells:
-            assert coeffs is not None
-            bases = [BSplineBasis(1, kts) for kts in patch.knots()]
-            shape = tuple(b.num_functions() for b in bases)
-            coeffs = splipy.utils.reshape(coeffs, shape, order='F')
-            patch = SplineObject(bases, coeffs, False, raw=True)
-            knots = [[(a+b)/2 for a, b in zip(t[:-1], t[1:])] for t in self.knots]
-            return patch(*knots)
-
-        if coeffs is not None:
-            coeffs = splipy.utils.reshape(coeffs, patch.shape, order='F')
-            if patch.rational:
-                coeffs = np.concatenate((coeffs, patch.controlpoints[..., -1, np.newaxis]), axis=-1)
-            patch = SplineObject(patch.bases, coeffs, patch.rational, raw=True)
-        return patch(*self.knots)
-
-    def elements(self):
-        nshape = tuple(len(k) for k in self.knots)
-        ranges = [range(k-1) for k in nshape]
-        nidxs = [np.array(q) for q in zip(*product(*ranges))]
-        eidxs = np.zeros((len(nidxs[0]), 2**len(nidxs)), dtype=int)
-        if len(nidxs) == 1:
-            eidxs[:,0] = nidxs[0]
-            eidxs[:,1] = nidxs[0] + 1
-        elif len(nidxs) == 2:
-            i, j = nidxs
-            eidxs[:,0] = np.ravel_multi_index((i, j), nshape)
-            eidxs[:,1] = np.ravel_multi_index((i+1, j), nshape)
-            eidxs[:,2] = np.ravel_multi_index((i+1, j+1), nshape)
-            eidxs[:,3] = np.ravel_multi_index((i, j+1), nshape)
-        elif len(nidxs) == 3:
-            i, j, k = nidxs
-            eidxs[:,0] = np.ravel_multi_index((i, j, k), nshape)
-            eidxs[:,1] = np.ravel_multi_index((i+1, j, k), nshape)
-            eidxs[:,2] = np.ravel_multi_index((i+1, j+1, k), nshape)
-            eidxs[:,3] = np.ravel_multi_index((i, j+1, k), nshape)
-            eidxs[:,4] = np.ravel_multi_index((i, j, k+1), nshape)
-            eidxs[:,5] = np.ravel_multi_index((i+1, j, k+1), nshape)
-            eidxs[:,6] = np.ravel_multi_index((i+1, j+1, k+1), nshape)
-            eidxs[:,7] = np.ravel_multi_index((i, j+1, k+1), nshape)
-
-        return len(nidxs), eidxs
-
-
-class LRSurfaceTesselation:
-
-    def __init__(self, patch):
-        nodes, elements = OrderedDict(), []
-        for el in patch.elements:
-            left, bottom = el.start()
-            right, top = el.end()
-
-            sw, se, nw, ne = (left, bottom), (right, bottom), (left, top), (right, top)
-            for pt in (sw, se, nw, ne):
-                nodes.setdefault(pt, len(nodes))
-
-            elements.append([nodes[sw], nodes[se], nodes[ne], nodes[nw]])
-
-        self._nodes = nodes
-        self._elements = np.array(elements, dtype=int)
-
-    def __call__(self, patch, coeffs=None, cells=False):
-        if cells:
-            assert coeffs is not None
-            ncomps = len(patch.elements) // coeffs.size
-            coeffs = coeffs.reshape((-1, ncomps))
-            return coeffs
-
-        if coeffs is not None:
-            patch = patch.clone()
-            patch.controlpoints = coeffs.reshape((len(patch.basis), -1))
-
-        return np.array([patch(*node) for node in self._nodes], dtype=float)
-
-    def elements(self):
-        return 2, self._elements
-
-
-class LRVolumeTesselation:
-
-    def __init__(self, patch):
-        nodes, elements = OrderedDict(), []
-        for el in patch.elements:
-            umin, vmin, wmin = el.start()
-            umax, vmax, wmax = el.end()
-
-            bsw, bse, bnw, bne = (umin, vmin, wmin), (umax, vmin, wmin), (umin, vmax, wmin), (umax, vmax, wmin)
-            tsw, tse, tnw, tne = (umin, vmin, wmax), (umax, vmin, wmax), (umin, vmax, wmax), (umax, vmax, wmax)
-            for pt in (bsw, bse, bnw, bne, tsw, tse, tnw, tne):
-                nodes.setdefault(pt, len(nodes))
-
-            elements.append([nodes[bsw], nodes[bse], nodes[bne], nodes[bnw],
-                             nodes[tsw], nodes[tse], nodes[tne], nodes[tnw]])
-
-        self._nodes = nodes
-        self._elements = np.array(elements, dtype=int)
-
-    def __call__(self, patch, coeffs=None, cells=False):
-        if cells:
-            assert coeffs is not None
-            ncomps = len(patch.elements) // coeffs.size
-            coeffs = coeffs.reshape((-1, ncomps))
-            return coeffs
-
-        if coeffs is not None:
-            patch = patch.clone()
-            patch.controlpoints = coeffs.reshape((len(patch.basis), -1))
-
-        return np.array([patch(*node) for node in self._nodes], dtype=float)
-
-    def elements(self):
-        return 3, self._elements
-
-
-def get_tesselation(patch):
-    if isinstance(patch, SplineObject):
-        return TensorTesselation(patch)
-    if isinstance(patch, lr.LRSplineSurface):
-        return LRSurfaceTesselation(patch)
-    if isinstance(patch, lr.LRSplineVolume):
-        return LRVolumeTesselation(patch)
-    assert False
+from .. import config
+from ..geometry import SplinePatch, LRPatch, UnstructuredPatch
 
 
 def expand_to_dims(array, dims=3):
@@ -214,14 +43,12 @@ class Basis:
         key = (stepid, patchid)
         if key not in self.patch_cache:
             g2bytes = self.reader.h5[str(stepid)][self.name]['basis'][str(patchid+1)][:].tobytes()
-            if g2bytes.startswith(b'# LRSPLINE SURFACE'):
-                patch = lr.LRSplineSurface(g2bytes)
-            elif g2bytes.startswith(b'# LRSPLINE VOLUME'):
-                patch = lr.LRSplineVolume(g2bytes)
+            if g2bytes.startswith(b'# LAGRANGIAN'):
+                patch = UnstructuredPatch.from_lagrangian(g2bytes)
+            elif g2bytes.startswith(b'# LRSPLINE'):
+                patch = LRPatch(g2bytes)
             else:
-                g2data = StringIO(g2bytes.decode())
-                with G2Object(g2data, 'r') as g:
-                    patch = g.read()[0]
+                patch = SplinePatch(g2bytes)
             self.patch_cache[key] = patch
         return self.patch_cache[key]
 
@@ -247,17 +74,12 @@ class Field:
         if self.ncomps is None:
             patch = self.basis.patch_at(stepid, 0)
             coeffs = self.coeffs(stepid, 0)
-            if not self.cells:
-                num, denom = len(coeffs), len(patch)
-            elif isinstance(patch, SplineObject):
-                ncells = np.prod([len(k)-1 for k in patch.knots()])
-                num, denom = len(coeffs), ncells
-            else:
-                ncells = len(patch.elements)
-                num, denom = len(coeffs), ncells
+
+            num = len(coeffs)
+            denom = patch.num_cells if self.cells else patch.num_nodes
 
             if num % denom != 0:
-                Log.critical('Inconsistent dimension in field "{}" ({}/{}); unable to discover number of components'.format(
+                log.error('Inconsistent dimension in field "{}" ({}/{}); unable to discover number of components'.format(
                     self.name, num, denom
                 ))
                 sys.exit(2)
@@ -269,14 +91,13 @@ class Field:
     def update(self, w, stepid):
         for patchid in range(self.basis.npatches):
             patch = self.basis.patch_at(stepid, patchid)
-            tess = self.reader.geometry.tesselation(patch)
             globpatchid = self.reader.geometry.findid(patch)
             if globpatchid is None:
                 continue
 
             coeffs = self.coeffs(stepid, patchid)
 
-            results = self.tesselate(patch, tess, coeffs)
+            results = self.tesselate(patch, coeffs)
             if hasattr(self, 'kind'):
                 kind = self.kind
             else:
@@ -294,8 +115,8 @@ class Field:
         sub = 'knotspan' if self.cells else 'fields'
         return self.reader.h5[str(stepid)][self.basis.name][sub][self.name][str(patchid+1)][:]
 
-    def tesselate(self, patch, tess, coeffs):
-        results = tess(patch, coeffs=coeffs, cells=self.cells)
+    def tesselate(self, patch, coeffs):
+        results = patch.tesselate_field(coeffs, cells=self.cells)
 
         if self.ncomps == 1 and self.vectorize:
             results = expand_to_dims(results)
@@ -338,10 +159,9 @@ class CombinedField(Field):
             results = []
             for src in self.sources:
                 patch = src.basis.patch_at(stepid, patchid)
-                tess = self.reader.geometry.tesselation(patch)
                 globpatchid = self.reader.geometry.findid(patch)
                 coeffs = src.coeffs(stepid, patchid)
-                results.append(src.tesselate(patch, tess, coeffs))
+                results.append(src.tesselate(patch, coeffs))
 
             if globpatchid is None:
                 continue
@@ -379,9 +199,9 @@ class SplitField(Field):
 
 class GeometryManager:
 
-    def __init__(self, basis, reader):
+    def __init__(self, basis):
         self.basis = basis
-        self.reader = reader
+        self.has_updated = False
 
         # Map knot vector -> evaluation points
         self.tesselations = {}
@@ -392,28 +212,21 @@ class GeometryManager:
         # Map corner tuple -> basisname x patchid
         self.corners = {}
 
-        Log.info('Using {} for geometry'.format(basis.name))
-
-    def tesselation(self, patch):
-        knots = tuple(tuple(k) for k in patch.knots())
-
-        if knots not in self.tesselations:
-            Log.debug('New unique knot vector detected, generating tesselation')
-            self.tesselations[knots] = get_tesselation(patch)
-        return self.tesselations[knots]
+        log.info('Using {} for geometry'.format(basis.name))
 
     def findid(self, patch):
-        corners = tuple(tuple(p) for p in patch.corners())
+        corners = patch.bounding_box
         if corners not in self.corners:
-            Log.error('Unable to find corresponding geometry patch')
+            log.error('Unable to find corresponding geometry patch')
             return None
         return self.globids[self.corners[corners]]
 
     def update(self, w, stepid):
-        if not self.basis.update_at(stepid):
+        if self.has_updated and not self.basis.update_at(stepid):
             return
+        self.has_updated = True
 
-        Log.info('Updating geometry')
+        log.info('Updating geometry')
 
         # FIXME: Here, all patches are updated whenever any of them are updated.
         # Maybe overkill.
@@ -423,23 +236,22 @@ class GeometryManager:
 
             if key not in self.globids:
                 self.globids[key] = len(self.globids)
-                Log.debug('New unique patch detected, generating global ID')
+                log.debug('New unique patch detected, generating global ID')
 
             # FIXME: This leaves behind invalidated corner IDs, which we should probably delete.
-            corners = tuple(tuple(p) for p in patch.corners())
-            self.corners[corners] = key
+            self.corners[patch.bounding_box] = key
             globpatchid = self.globids[(self.basis.name, patchid)]
 
-            tess = self.tesselation(patch)
-            nodes = tess(patch)
+            tess = patch.tesselate()
+            nodes = tess.nodes
+            dim = 2 if tess.cells.shape[-1] == 4 else 3
+            eidxs = tess.cells
 
             while nodes.shape[-1] < 3:
                 nshape = nodes.shape[:-1] + (1,)
                 nodes = np.concatenate((nodes, np.zeros(nshape)), axis=-1)
 
-            dim, eidxs = tess.elements()
-
-            Log.debug('Writing patch {}'.format(globpatchid))
+            log.debug('Writing patch {}'.format(globpatchid))
             w.update_geometry(nodes, eidxs, dim, globpatchid)
 
         w.finalize_geometry(stepid)
@@ -447,10 +259,8 @@ class GeometryManager:
 
 class Reader:
 
-    def __init__(self, h5, bases=(), geometry=None):
+    def __init__(self, h5):
         self.h5 = h5
-        self.only_bases = bases
-        self.geometry_basis = geometry
 
     def __enter__(self):
         self.bases = OrderedDict()
@@ -464,10 +274,10 @@ class Reader:
         self.combine_fields()
         self.sort_fields()
 
-        if self.geometry_basis:
-            self.geometry = GeometryManager(self.bases[self.geometry_basis], self)
+        if config.geometry_basis:
+            self.geometry = GeometryManager(self.bases[config.geometry_basis])
         else:
-            self.geometry = GeometryManager(next(iter(self.bases.values())), self)
+            self.geometry = GeometryManager(next(iter(self.bases.values())))
 
         return self
 
@@ -480,20 +290,28 @@ class Reader:
 
     def steps(self):
         for stepid in range(self.nsteps):
-            # FIXME: Grab actual time here as second element
-            yield stepid, {'time': float(stepid)}, self.h5[str(stepid)]
+            try:
+                time = self.h5[str(stepid)]['timeinfo']['level'][0]
+            except KeyError:
+                time = float(stepid)
+            yield stepid, {'time': time}, self.h5[str(stepid)]
 
     def outputsteps(self):
-        for stepid, time, _ in self.steps():
+        if config.only_final_timestep:
+            for stepid, time, _ in self.steps():
+                pass
             yield stepid, time
+        else:
+            for stepid, time, _ in self.steps():
+                yield stepid, time
 
     def allowed_bases(self, group, items=False):
-        if not self.only_bases:
+        if not config.only_bases:
             yield from (group.items() if items else group)
         elif items:
-            yield from ((b, v) for b, v in group.items() if b in self.only_bases)
+            yield from ((b, v) for b, v in group.items() if b in config.only_bases)
         else:
-            yield from (b for b in group if b in self.only_bases)
+            yield from (b for b in group if b in config.only_bases)
 
     def init_bases(self):
         for stepid, _, stepgrp in self.steps():
@@ -513,19 +331,20 @@ class Reader:
         # Delete bases that don't have any patch data
         to_del = [name for name, basis in self.bases.items() if not basis.update_steps]
         for basisname in to_del:
-            Log.debug('Basis {} has no updates, removed'.format(basisname))
+            log.debug('Basis {} has no updates, removed'.format(basisname))
             del self.bases[basisname]
 
         # Delete the bases we don't need
-        if self.only_bases:
-            self.only_bases = set(self.bases) & self.only_bases
-            keep = self.only_bases + ((self.geometry_basis,) if self.geometry_basis else ())
+        # TODO: Don't do it this way, though
+        if config.only_bases:
+            config.require(only_bases=tuple(set(self.bases) & set(config.only_bases)))
+            keep = config.only_bases + ((config.geometry_basis,) if config.geometry_basis else ())
             self.bases = OrderedDict((b,v) for b,v in self.bases.items() if b in keep)
         else:
-            self.only_bases = set(self.bases)
+            config.require(basis=tuple(self.bases.keys()))
 
         for basis in self.bases.values():
-            Log.debug('Basis {} updates at steps {} ({} patches)'.format(
+            log.debug('Basis {} updates at steps {} ({} patches)'.format(
                 basis.name, ', '.join(str(s) for s in sorted(basis.update_steps)), basis.npatches,
             ))
 
@@ -536,17 +355,19 @@ class Reader:
                 kspans = basisgrp['knotspan'] if 'knotspan' in basisgrp else []
 
                 for fieldname in fields:
-                    if fieldname not in self.fields:
+                    if fieldname not in self.fields and basisname in self.bases:
                         self.fields[fieldname] = Field(fieldname, self.bases[basisname], self)
-                    self.fields[fieldname].add_update(stepid)
+                    if fieldname in self.fields:
+                        self.fields[fieldname].add_update(stepid)
                 for fieldname in kspans:
-                    if fieldname not in self.fields:
+                    if fieldname not in self.fields and basisname in self.bases:
                         self.fields[fieldname] = Field(fieldname, self.bases[basisname], self, cells=True)
-                    self.fields[fieldname].add_update(stepid)
+                    if fieldname in self.fields:
+                        self.fields[fieldname].add_update(stepid)
 
     def log_fields(self):
         for field in self.fields.values():
-            Log.debug('{} "{}" lives on {} ({} components)'.format(
+            log.debug('{} "{}" lives on {} ({} components)'.format(
                 'Knotspan' if field.cells else 'Field', field.name, field.basis.name, field.ncomps
             ))
 
@@ -563,9 +384,9 @@ class Reader:
                     splitnames = ['{} {}'.format(prefix, name) for name in splitnames]
 
                 if any(splitname in self.fields for splitname in splitnames):
-                    Log.warning('Unable to split "{}", some fields already exist'.format(fname))
+                    log.warning('Unable to split "{}", some fields already exist'.format(fname))
                     continue
-                Log.info('Split field "{}" -> {}'.format(fname, ', '.join(splitnames)))
+                log.info('Split field "{}" -> {}'.format(fname, ', '.join(splitnames)))
                 for i, splitname in enumerate(splitnames):
                     to_add.append(SplitField(splitname, self.fields[fname], i, self))
                 to_del.append(fname)
@@ -590,9 +411,9 @@ class Reader:
             sources = [self.fields[s] for s in sourcenames]
             try:
                 self.fields[fname] = CombinedField(fname, sources, self)
-                Log.info('Creating combined field {} -> {} ({} components)'.format(', '.join(sourcenames), fname, len(sources)))
+                log.info('Creating combined field {} -> {} ({} components)'.format(', '.join(sourcenames), fname, len(sources)))
             except AssertionError:
-                Log.warning('Unable to combine {} -> {}'.format(', '.join(sourcenames), fname))
+                log.warning('Unable to combine {} -> {}'.format(', '.join(sourcenames), fname))
 
     def sort_fields(self):
         fields = sorted(self.fields.values(), key=lambda f: f.name)
@@ -600,19 +421,16 @@ class Reader:
         self.fields = OrderedDict((f.name, f) for f in fields)
 
     def write(self, w):
-        for stepid, time in self.outputsteps():
-            Log.info('Step {}'.format(stepid))
+        for stepid, time in log.iter.plain('Step', self.outputsteps()):
+            w.add_step(**time)
+            self.geometry.update(w, stepid)
 
-            with Log.indent():
-                w.add_step(**time)
-                self.geometry.update(w, stepid)
+            for field in self.fields.values():
+                if field.update_at(stepid):
+                    log.info('Updating {} ({})'.format(field.name, field.basisname))
+                    field.update(w, stepid)
 
-                for field in self.fields.values():
-                    if field.update_at(stepid):
-                        Log.info('Updating {} ({})'.format(field.name, field.basisname))
-                        field.update(w, stepid)
-
-                w.finalize_step()
+            w.finalize_step()
 
     def write_mode(self, w, mid, field):
         for pid in range(self.npatches(0, field.basis)):
@@ -686,10 +504,10 @@ class EigenReader(Reader):
         self.fields['Mode Shape'] = field
 
 
-def get_reader(filename, bases=(), geometry=None, **kwargs):
+def get_reader(filename):
     h5 = h5py.File(filename, 'r')
     basisname = next(iter(h5['0']))
     if 'Eigenmode' in h5['0'][basisname]:
-        Log.info('Detected eigenmodes')
-        return EigenReader(h5, bases=bases, geometry=geometry)
-    return Reader(h5, bases=bases, geometry=geometry)
+        log.info('Detected eigenmodes')
+        return EigenReader(h5)
+    return Reader(h5)
