@@ -1,6 +1,8 @@
 import lrspline
 
-from .hdf5 import Log, GeometryManager
+from .. import config
+from ..geometry import LRPatch
+from .hdf5 import GeometryManager
 
 
 class SimpleBasis:
@@ -22,13 +24,13 @@ class SimpleBasis:
 
 class Reader:
 
-    def __init__(self, filename, **kwargs):
-        self.basis = None
+    def __init__(self, filename):
         self.filename = filename
+        config.require(multiple_timesteps=False)
 
     def __enter__(self):
         with open(self.filename, 'rb') as f:
-            patches = lrspline.LRSplineObject.read_many(f)
+            patches = list(map(LRPatch, lrspline.LRSplineObject.read_many(f)))
         self.basis = SimpleBasis(patches)
         return self
 
@@ -37,6 +39,6 @@ class Reader:
 
     def write(self, w):
         w.add_step(time=0.0)
-        geometry = GeometryManager(self.basis, self)
+        geometry = GeometryManager(self.basis)
         geometry.update(w, 0)
         w.finalize_step()
